// ZEVA 8-cell Battery Monitor Module V2
// Open Source version, released under MIT License (see Readme file)
// Last modified by Ian Hooper (ZEVA), August 2021

// Code for ATmega48PA
// Fuses: Leave as defaults

#define F_CPU 1000000UL

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/wdt.h>
#include <avr/sleep.h>

// Defining voltage thresholds. Measurement sensitivity is 37.4mV/ADC
// So normal values for LiFePO4 with hysteresis are 66 (2.5V), 80 (3.0V), 102 (3.8V), 91 (3.4V)
// LiCo / NMC can be 80 (3.0V), 85 (3.2V), 111 (4.2V), 107 (4.0V)
#define	LV_OFF_THRESHOLD	80
#define LV_ON_THRESHOLD		88
#define	HV_OFF_THRESHOLD	99
#define HV_ON_THRESHOLD		91

#define LATCHING	0 // This prevents outputs being re-enabled if they have tripped

#define VERY_LOW_THRESHOLD	40 // Point at which module turns itself off - normally 40 = about 1.5V

// I/O defines
#define HV_RELAY	(1<<PC4)
#define	LV_RELAY	(1<<PC3)
#define PWR_HOLD	(1<<PC0)
#define GREEN		(1<<PB0)
#define RED			(1<<PB1)

// Function declarations
short GetCell(int n); // 0-8
short ReadADC(unsigned char adc_input);

// Global variable
short level[9][8];

void SampleCells(int counter)
{
	for (int n=1; n<=8; n++)
		level[n][counter] = GetCell(n);
	GetCell(0); // Turn sampling off
}

int main(void)
{
	// Setup Ports
    DDRB = 0b11000011; // PB6-7 are cell samplers 7-8, PB0-1 are green and red LEDs
    DDRC = 0b00111001; // PC3: LVOUT, PC4: HV_OUT, PC5: cell 1 sampler, PC0: PWR_HOLD
    DDRD = 0b00011111; // PD0-4 for cell samplers 2-6
    ADCSRA = 0b10000011; // Top bit turns ADC on, the 011 LSBs gives a division factor of 8 (1mhz / 8 -> 125kHz ADC clock)

    PORTC |= PWR_HOLD;

	_delay_ms(500); // Let things settle

	char lows[9];
	char highs[9];
	char low = 0;
	char high = 0;
	char veryLows[9];
	char veryLow = 0;

	for (int n=0; n<9; n++) lows[n] = highs[n] = veryLows[n] = 0; // Make sure they all start at zero

	// Detect number of cells and flash LED that many times for confirmation
	for (int n=0; n<8; n++) level[0][n] = 0;
	SampleCells(0);
	int cellCount;
	for (cellCount=1; cellCount<=8; cellCount++)
	{
		int diff = level[cellCount][0] - level[cellCount-1][0];
		if (diff < LV_OFF_THRESHOLD/2) break;
	}
	cellCount--;
	
	if (cellCount == 0) // No cells detected (probably some fault/error)
	{
		while(1) // Flash LEDs alternately forever
		{
			PORTB = GREEN;
			_delay_ms(400);
			PORTB = RED;
			_delay_ms(400);
		}
	}

	for (int n=1; n<=cellCount; n++) // Flash LEDs according to number of cells
	{	
		PORTB |= GREEN;
		_delay_ms(150);
		PORTB &= ~GREEN;
		_delay_ms(150);
	}

	// Pre-fill array with current voltage levels
	for (int n=0; n<8; n++) SampleCells(n);

	_delay_ms(500);
	unsigned short counter = 0;
	while (1)
	{
		counter++;
		if (counter == 8) counter = 0;

		SampleCells(counter);

		int avgLevel[9];
		avgLevel[0] = 0;
		for (int c=1; c<=cellCount; c++)
		{
			avgLevel[c] = 0;
			for (int n=0; n<8; n++)
				avgLevel[c] += level[c][n];
			avgLevel[c] /= 8;
		}

		for (int n=1; n<=cellCount; n++)
		{
			int diff = avgLevel[n] - avgLevel[n-1];
			
			// Tiny bit of hysteresis around the threshold value
			if (diff < LV_OFF_THRESHOLD && lows[n] < 100)
				lows[n]++;
			else if (diff > LV_ON_THRESHOLD)
				lows[n] = 0;

			if (diff > HV_OFF_THRESHOLD && highs[n] < 100)
				highs[n]++;
			else if (diff < HV_ON_THRESHOLD)
				highs[n] = 0;

			if (diff < VERY_LOW_THRESHOLD)
				veryLows[n]++;
			else
				veryLows[n] = 0;
		}

		if (!LATCHING) // Reset low & high signals if not latching
		{
			low = 0;
			high = 0;
		}
		
		for (int n=1; n<=cellCount; n++)
		{
			if (lows[n] > 5) low = 1;
			if (highs[n] > 5) high = 1;
			if (veryLows[n] > 5) veryLow = 1;
		}

		// Reverse logic, so output is closed whenever all modules are happy
		if (low)
			PORTC &= ~LV_RELAY;
		else
			PORTC |= LV_RELAY;

		if (high)
			PORTC &= ~HV_RELAY;
		else
			PORTC |= HV_RELAY;

		if (veryLow) PORTC &= ~PWR_HOLD; // Power off if any cells getting extremely low

		if (!low && !high)
		{
			PORTB &= ~RED;
			PORTB |= GREEN;
		}
		else
		{
			if (low) // Blink red
			{
				PORTB &= ~GREEN;
				if (counter == 1 || counter == 5)
					PORTB |= RED;
				else
					PORTB &= ~RED;
			}
			else // Solid red
			{
				PORTB |= RED;
				PORTB &= ~GREEN;
			}
		}

		_delay_ms(250); // slow iterations down a little, slightly dirty coding
	}
}


short GetCell(int n)
{
	// Turn all HV sampler pins off
	PORTB &= 0b00111111;
	PORTC &= 0b11011111;
	PORTD &= 0b11100000;

	// According to datasheet, FOD852 can take 100us to turn off and up to 300us to turn on
	// Plus we add a bit just for margin

	//_delay_us(300); // don't shorten these delays or it breaks
	_delay_us(800); // Trying much longer to really rule out that overlap situation
	
	switch (n) // Turn selected sampler on
	{
		case 1: PORTC |= (1<<PC5); break;
		case 2: PORTD |= (1<<PD0); break;
		case 3: PORTD |= (1<<PD1); break;
		case 4: PORTD |= (1<<PD2); break;
		case 5: PORTD |= (1<<PD3); break;
		case 6: PORTD |= (1<<PD4); break;
		case 7: PORTB |= (1<<PB6); break;
		case 8: PORTB |= (1<<PB7); break;
		default: return 0;
	}

	_delay_us(200); // Let capacitor stabilise
	return ReadADC(1) + 18; // To compensate for optocoupler voltage drop of about 0.6V (effects first cell measurement)
	// Was +15, but cell 1 was reading lower than it should, i.e tripping too high
}

short ReadADC(unsigned char adc_input)
{
	ADMUX = (1<<REFS0) + adc_input; // REFS0 means AVcc reference with ext capacitor on AREF
	_delay_us(10); // Delay to allow for changing of ADC system (single vs differential)
	ADCSRA |= 0x40; // Start the AD conversion
	while ((ADCSRA & 0x10) == 0); // Wait for the AD conversion to complete
	ADCSRA |= 0x10;
	return ADCW;
}
